##
## This file contains library proc's that are used in many places
## by the tkined scripts. They are auto loaded using tcl's autoload
## facility.
##
## Copyright (c) 1993, 1994
##
## J. Schoenwaelder
## TU Braunschweig, Germany
## Institute for Operating Systems and Computer Networks
##
## Permission to use, copy, modify, and distribute this
## software and its documentation for any purpose and without
## fee is hereby granted, provided that this copyright
## notice appears in all copies.  The University of Braunschweig
## makes no representations about the suitability of this
## software for any purpose.  It is provided "as is" without
## express or implied warranty.
##


##
## =========================================================================
## =================== T K I N E D   related subroutines ===================
## =========================================================================
##
##  LoadDefaults { class }
##  ShowDefaults
##
##  write   { {txt ""} {cmd {}}}
##  writeln { {txt ""} {cmd {}}}
##  debug   { txt }
##
##  ldelete { list elem }
##
## =========================================================================
##

##
## Search for tkined.defaults files following the auto_path and read
## default definition. Each default definition has the following syntax:
##
##    <class>.<attribute>.<index>: <value>
##
## Every entry matching the class argument will be written to the global
## default array, indexed by <attribute>.<index>. A sample application
## is the definition of snmp related parameters like:
##
##    snmp.port.134.169.34.15: 8132
##    snmp.community.134.169.34.15: private
##
## The resulting tcl variables are:
##
##    default(port.134.169.34.15) -> 8132
##    default(community.134.169.34.15) -> private
##

proc LoadDefaults { args } {

    global auto_path 
    global default

    set filename tkined.defaults
    
    set reverse_path ""
    foreach dir $auto_path {
	set reverse_path "$dir $reverse_path"
    }
    
    foreach dir $reverse_path {
	if [file readable $dir/$filename] {
	    set fh [open $dir/$filename r]
	    while {![eof $fh]} {
		gets $fh line
		set line [string trim $line]
		if {($line == "") || ([regexp "^#|^!" $line])} continue
		foreach class $args {
		    if {[string match $class.* $line]} {
			set line [split $line ":"]
			set list [split [lindex $line 0] .]
			set name  [join [lrange $list 1 end] .]
			set value [string trim [lindex $line 1]]
			set default($name) $value
		    }
		}
	    }
	    close $fh
	}
    }
    return ""
}

##
## Show the defaults as loaded from the tkined.defaults files.
##

proc ShowDefaults {} {

    global default

    if {![info exists default]} {
	ined acknowledge \
	    "No parameters used from the tkined.default files."
	return
    }

    set result ""
    foreach name [array names default] {
	set attname  [lindex [split $name .] 0]
	set attindex [join [lrange [split $name .] 1 end] .]
	lappend result [format "%-16s %-16s %s" \
			$attname $attindex $default($name)]
    }

    set result [lsort $result]
    ined browse "Defined defaults used by this application:" "" \
	"Attribute         Index            Value" $result
}

##
## Write a report to a log object. The name is expected to be in
## the variable tool_name which may be set in one of the init
## procs below.
##

proc writeln {{txt ""} {cmd ""}} {
    write "$txt\n" $cmd
}

proc write {{txt ""} {cmd ""}} {
    static log
    global tool_name

    if {![info exists tool_name]} {
	set tool_name "Unknown"
    }

    if {(![info exists log]) || ([ined -noupdate retrieve $log] == "")} {
	set log [ined -noupdate create LOG]
	ined -noupdate name $log "$tool_name Report"
    }

    if {$cmd == ""} {
	ined append $log "$txt"
    } else {
	ined hyperlink $log $cmd $txt
    }
}

##
## Print messages if running in debug mode.
##

proc debug {args} {

    global debug

    if {[info exists debug] && ($debug == "true")} {
	foreach arg $args {
	    writeln $arg
	}
    }
}

##
## Delete an element from a list.
##

proc ldelete {list element} {
    upvar $list mylist
    set result ""
    foreach e $mylist {
        if {$e != $element} { lappend result $e }
    }
    return $result
}


##
## =========================================================================
## ================== M O N I T O R   related subroutines ==================
## =========================================================================
##
##  CloneNode { id clone {offset_x 20} {offset_y 20} }
##  DeleteClones {}
##  MoJoCheckIds { cmd ids } 
##  MoJoInfo {}
##  MoJoSelect { action } 
##  MoJoModify {}
##
## =========================================================================
##

##
## Clone a node object given by id and move the new object to a
## position given by relative coordinates offset_x and offset_y.
## The clone operation is only done when the global variable
## clone is set to true. All new object ids are collected in the
## global list clone_ids.
##

proc CloneNode { id clone {offset_x 20} {offset_y 20} } {

    global clone_ids

    if {![info exists clone_ids]} { set clone_ids "" }

    ined -noupdate name    $clone [ined -noupdate name $id]
    ined -noupdate address $clone [ined -noupdate address $id]
    ined -noupdate color   $clone [ined -noupdate color $id]
    ined -noupdate font    $clone [ined -noupdate font $id]
    ined -noupdate label   $clone [ined -noupdate label $id]
    ined -noupdate icon    $clone [ined -noupdate icon $id]
    set xy [ined -noupdate move $id]
    set x [expr {[lindex $xy 0]+$offset_x}]
    set y [expr {[lindex $xy 1]+$offset_y}]
    ined -noupdate move $clone $x $y

    lappend clone_ids $clone

    return $clone
}

##
## Delete all previously created clones.
##

proc DeleteClones {} {

    global clone_ids

    if {[info exists clone_ids]} {
	foreach id $clone_ids {
	    ined delete $id
	}
    }
}

##
## Check if all ined objects given by ids are still there. Modify the 
## current job if any objects have been removed and kill the job, if no 
## ids have been left. This proc assumes that a command is simply a 
## command name and a list of ids as arguments.
##

proc MoJoCheckIds { cmd ids } {

    set new_ids ""
    foreach id $ids {
        if {[ined retrieve $id] != ""} {
            lappend new_ids $id
        }
    }

    if {$new_ids != $ids} {
        if {$new_ids == ""} {
            job kill [job current]
	    return 
        } else {
	    job command [job current] [list $cmd $new_ids]
        }
    }

    return $new_ids
}

##
## Display the jobs currently running. The ids of the command are
## converted to the name attributes to increase readability.
##

proc MoJoInfo { } {

    set jobs [job list]

    if {$jobs == ""} {
	ined acknowledge "Sorry, no jobs available."
	return
    }
    
    set result ""
    set len 0
    foreach j $jobs {

	set jobid   [lindex $j 0]
	set jobcmd  [lindex $j 1]
	set jobitv  [expr {[lindex $j 2] / 1000.0}]
	set jobrem  [expr {[lindex $j 3] / 1000.0}]
	set jobcnt  [lindex $j 4]
	set jobstat [lindex $j 5]

	set line \
	     [format "%s %6.1f %6.1f %3d %8s %s" \
	      $jobid $jobitv $jobrem $jobcnt $jobstat \
	      [lindex $jobcmd 0] ]

	# Convert the id's to hostnames for readability.
	
	foreach id [lindex $jobcmd 1] {
	    set host [lindex [ined name $id] 0]
	    if {[string length $line] < 65} {
		append line " $host"
	    } else {
		lappend result $line
		set line [format "%35s %s" "" $host]
	    }
	}

	lappend result $line
    }

    set header " ID    INTV    REM  CNT STATUS  COMMAND"

    foreach line $result {
	if {[string length $line] > $len} {
	    set len [string length $line]
	}
    }

    for {set i [string length $header]} {$i < $len} {incr i} {
	append header " "
    }

    ined browse $header $result
}

##
## Select one of the existing jobs for doing action with it.
## This is at least used by the MoJoModify proc below.
##

proc MoJoSelect { action } {
    
    set jobs [job list]

    if {$jobs == ""} {
	ined acknowledge "Sorry, no job to $action."
        return
    }

    set res ""
    foreach j $jobs {
	# Convert the id's to hostnames for readability.
	
	set hosts [lindex [lindex $j 1] 0]
	foreach id [lindex [lindex $j 1] 1] {
	    lappend hosts [lindex [ined -noupdate name $id] 0]
	}
	if {[llength $hosts] > 4} {
	    set hosts "[lrange $hosts 0 3] ..."
	}
    
	set status [lindex $j 5]
	if {   ($action == "resume"  && $status == "suspend")
            || ($action == "suspend" && $status == "waiting")
            || ( ($action != "resume") && ($action != "suspend"))} {
		lappend res [format "%s %s" [lindex $j 0] $hosts]
	    }
    }

    if {$res == ""} {
	ined acknowledge "Sorry, no job to $action."
	return ""
    } 

    set res [ined list "Choose a job to $action:" $res [list cancel $action]]
    if {[lindex $res 0] == "cancel"} {
	return ""
    } else {
	return [lindex [lindex $res 1] 0]
    }
}

proc MoJoModify {} {

    static threshold

    # Ask for the job to modify.

    set jobid [MoJoSelect modify]

    if {$jobid == ""} return

    # Set up default values.

    if {![info exists threshold($jobid)]} {
	set threshold($jobid) ""
    }

    # Get the details about the selected job.

    set jobcmd ""
    set jobitv ""
    foreach job [job list] {
	if {$jobid == [lindex $job 0]} {
	    set jobcmd  [lindex $job 1]
	    set jobitv  [expr {[lindex $job 2] / 1000.0}]
	    set jobstat [lindex $job 5]
	}
    }
    
    # Convert the status waiting to active
    
    if {$jobstat == "waiting"} { set jobstat active }
	
    # Convert the id's to hostnames for readability.
    
    set hosts ""
    foreach id [lindex $jobcmd 1] {
	lappend hosts [lindex [ined -noupdate name $id] 0]
    }
    if {[llength $hosts] > 3} {
	set hosts "[lrange $hosts 0 2] ..."
    }
    
    # Request for changes.
    
    set value $threshold($jobid)

    set res [ined request "Modify $jobid ([lindex $jobcmd 0] $hosts)" \
	      [list [list "Intervaltime \[s\]:" $jobitv entry 10] \
	            [list "Job Status:" $jobstat radio active suspend] \
		    [list "Threshold:" $value entry 10] ] \
	      [list cancel "kill job" modify] ]
    
    if {[lindex $res 0] == "cancel"} return

    if {[lindex $res 0] == "kill job"} {
	job kill $jobid
	return
    }
    
    set jobitv  [lindex $res 1]
    set jobstat [lindex $res 2]
    set value   [lindex $res 3]

    set threshold($jobid) $value
    
    if {$jobstat == "active"}  { job status $jobid waiting }
    if {$jobstat == "suspend"} { job status $jobid suspend }

    if {[catch {expr round($jobitv * 1000)} ms] || $jobitv <= 0} {
	ined acknowledge "Illegal interval time ignored."
	return
    }

    if {$ms != [job interval $jobid]} {

	job interval $jobid $ms

	set newlist ""
	foreach cmd [ined restart] {
	    if {[lindex $cmd 3] == $jobid} {
		set cmd [lreplace $cmd 2 2 $jobitv]
	    }
	    lappend newlist $cmd
	}
	ined restart $newlist
    }

    # Now add, update or remove the threshold attribute.

    foreach id [lindex $jobcmd 1] {
	ined attribute $id threshold $value
    }
}

proc MoJoCheckThreshold { id txt value {unit ""} } {

    set action [MoJoThresholdAction]
    if {$action == ""} return

    set threshold [ined attribute $id threshold]
    if {$threshold == ""} return

    set len [llength $value]
    for {set i 0} {$i < $len} {incr i} {
	catch {
	    set v [lindex $value $i]
	    set t [lindex $threshold $i]
	    if {$t == ""} continue
	    if {$v > $t} {
		set msg "$txt $v $unit exceeds threshold $t $unit"
		if {[lsearch $action syslog] >= 0} {
		    syslog warning $msg
		}
		if {[lsearch $action flash] >= 0} {
		    set jobid [job current]
		    if {$jobid != ""} {
			set secs [expr {[job interval $jobid] / 1000}]
		    } else {
			set secs 2
		    }
		    ined flash $id $secs
		}
		if {[lsearch $action write] >= 0} {
		    writeln "[getdate]:"
		    writeln $msg
		    writeln
                }
	    }
	}
    }
}

proc MoJoThresholdAction { args } {
    static action
    if {![info exists action]} { set action [list flash write] }
    if {$args != ""} { set action [lindex $args 0] }
    return $action
}


##
## =========================================================================
## ======================= I P   related subroutines =======================
## =========================================================================
##
##  IpInit { toolname }
##
##  GetIpAddress { node }
##  IpFlash { ip }
##  IpService {}
##
## =========================================================================
##

##
## Initialize global variables that are used by these ip procs.
##

proc IpInit { toolname } {

    global default icmp_retries icmp_timeout icmp_delay icmp_routelength
    global tool_name

    set tool_name $toolname

    if {[info exists default(retries)]} {
	set icmp_retries $default(retries)
    } else {
	set icmp_retries 3
    }
    icmp -retries $icmp_retries
    
    if {[info exists default(timeout)]} {
	set icmp_timeout $default(timeout)
    } else {
	set icmp_timeout 3
    }
    icmp -timeout $icmp_timeout
    
    if {[info exists default(delay)]} {
	set icmp_delay $default(delay)
    } else {
	set icmp_delay 5
    }
    icmp -delay $icmp_delay
    
    if {[info exists default(routelength)]} {
	set icmp_routelength $default(routelength)
    } else {
	set icmp_routelength 16
    }
}

##
## Get the IP Address of a node. Query the name server, if the
## address attribute is not set to something that looks like a
## valid IP address.
##

proc GetIpAddress { node } {
    if {[lsearch "NODE STRIPCHART BARCHART" [ined type $node]] >= 0} {
        set host [ined name $node]
        set ip [ined address $node]
        if {[regexp "^\[0-9\]+\.\[0-9\]+\.\[0-9\]+\.\[0-9\]+$" $ip] > 0} {
            return $ip
        }
	if {[regexp "^\[0-9\]+\.\[0-9\]+\.\[0-9\]+\.\[0-9\]+$" $host] > 0} {
	    return $host
	}
        if {[catch {nslook $host} ip]==0} {
            return [lindex $ip 0]
        }
    }
    return ""
}

##
## Flash the icon of the object given by the ip address.
##

proc IpFlash { ip {secs 2} } {
    foreach comp [ined retrieve] {
	if {([ined type $comp] == "NODE") && ([GetIpAddress $comp] == $ip)} {
	    ined flash [ined id $comp] $secs
	}
    }
}

##
## Read the file /etc/services and let the user select a TCP service of
## interest. The proc returns a list containing the name and the port number.
##

proc IpService { protocol } {

    if {[catch {open /etc/services} fh]} {
	ined acknowledge "Can not open /etc/services."
        return
    }
    
    set service(X11) 6000
    while {! [eof $fh]} {
	gets $fh line
	if {[string index [string trim $line] 0] == "#"} continue
	if {[scan $line "%s %d/$protocol" name number] != 2} continue
	set service($name) $number
    }
    close $fh

    set result [ined list "Select the TCP service of interest:" \
		  [lsort [array names service]] \
		  [list cancel define select]]

    switch [lindex $result 0] {
	cancel { return }
	select {
	    set name [lindex [lindex $result 1] 0]
	    if {$name == ""} return
	}
	define {
	    static sname
	    static sport
	    if {![info exists sport]} { set sport "" }
	    if {![info exists sname]} { set sname "noname" }
	    set result [ined request "Set name and port number for service:" \
			 [list [list Name: $sname] [list Port: $sport]] \
			 [list cancel ok] ]
	    if {[lindex $result 0] == "cancel"} return
	    set sname [lindex $result 1]
	    set sport [lindex $result 2]
	    set name $sname
	    set service($name) $sport
	}
	{} {
	    set name [lindex [lindex $result 1] 0]
            if {$name == ""} return
	}
    }

    return [list $name $service($name)]
}

##
## =========================================================================
## ========== S N M P   related subroutines (based on PHK's snmp) ==========
## =========================================================================
##
##  SnmpInit { toolname }
##  SnmpParameter
##  SnmpOpen  { ip }
##  SnmpGet { s varlist }
##  SnmpGetBulk { s varlist varname body }
##  SnmpClose { s }
##  SnmpWalk  { ip path }
##
## =========================================================================
##


##
## Initialize global variables that are used by these snmp procs.
##

proc SnmpInit { toolname } {

    global snmp_community snmp_timeout snmp_retries snmp_port snmp_verbose 
    global snmp_browser
    global tool_name
    global default

    if {[info exists default(community)]} {
	set snmp_community $default(community)
    } else {
	set snmp_community "public"
    }

    if {[info exists default(timeout)]} {
	set snmp_timeout $default(timeout)
    } else {
	set snmp_timeout 5000000
    }

    if {[info exists default(retries)]} {
	set snmp_retries $default(retries)
    } else {
	set snmp_retries 3
    }

    if {[info exists default(port)]} {
	set snmp_port $default(port)
    } else {
	set snmp_port 161
    }

    set snmp_verbose true
    set snmp_browser ""

    set tool_name $toolname

    if {[lsearch [info commands] snmp] < 0} {
	ined acknowledge "Sorry, this version of scotty has no snmp support."
	exit
    }
}


##
## Set the parameters (community, timeout, retry) for snmp requests.
##

proc SnmpParameter {} {

    global snmp_community snmp_timeout snmp_retries snmp_port snmp_verbose

    set result [ined request "SNMP Parameter" \
		[list [list "Community:" $snmp_community entry 20] \
                      [list "UDP Port:" $snmp_port entry 10] \
                      [list "Timeout \[us\]:" $snmp_timeout entry 10] \
                      [list "Retries:" $snmp_retries scale 1 8] \
                      [list "Verbose:" $snmp_verbose radio true false] ] \
		[list cancel "set values"] ]

    if {[lindex $result 0] == "cancel"} return

    set snmp_community [lindex $result 1]
    set snmp_port      [lindex $result 2]
    set snmp_timeout   [lindex $result 3]
    set snmp_retries   [lindex $result 4]
    set snmp_verbose   [lindex $result 5]
}

##
## Open a snmp session and initialize it to the defaults stored in
## the global variables snmp_community, snmp_timeout, snmp_retries.
##

proc SnmpOpen { ip } {

    global snmp_community snmp_timeout snmp_retries snmp_port default

    set s [snmp open $ip]

    snmp default $s community $snmp_community
    snmp default $s timeout $snmp_timeout
    snmp default $s retry $snmp_retries
    catch {snmp default $s port $snmp_port}

    if {[info exists default(community.$ip)]} { 
	snmp default $s community $default(community.$ip) 
    }
    if {[info exists default(timeout.$ip)]} { 
	snmp default $s timeout $default(timeout.$ip) 
    }
    if {[info exists default(retries.$ip)]} { 
	snmp default $s retry $default(retries.$ip) 
    }
    if {[info exists default(port.$ip)]} {
	catch {snmp default $s port $default(port.$ip)}
    }

    return $s
}

##
## Close an snmp handle. The catch is neede to get around a bug in
## PHK's interface.
##

proc SnmpClose { s } {
    catch {snmp close $s}
}

##
## Send a snmp get request and return the retrieves values.
##

proc SnmpGet { s varlist } {
    return [snmp get $s -l -v $varlist]
}

##
## Send a snmp getnext request and set the variable varname to the
## retrieved values. Evaluate the body foreach getnext that is in
## our subtree.
##

proc SnmpGetBulk { s varlist varname body } {
    uplevel 1 [list snmp getbulk $s -l -v $varlist $varname $body]
}

##
## Dump a hierarchy of the MIB. Sometimes called a walk in SNMP
## terminology.
##

proc SnmpWalk {ip path} {

    global snmp_verbose

    if {[catch {nslook $ip} hostname]} {
	set hostname ""
    }

    if {$snmp_verbose != "true"} {
	writeln "$hostname \[$ip\]\n$path\n"
    }

    set s [SnmpOpen $ip]
    set op ""
    set count 0
    if {[catch {
	snmp getbulk $s {$path} x {

	    set oid [lindex $x 0]
	    set typ [lindex $x 1]
	    set val [lindex $x 2]

	    # get the oid from the MIB
	    set l [mib oid $oid]
	    set p [lindex $l 0]
	    set pl [split $p .]
            set varname [lindex $pl [expr {[llength $pl]-1}]]

	    if {$p == $op} {
		writeln [format "\t%-16s = %s" \
			 "$varname\[[lindex $l 1]\]" $val]
		continue
	    }

	    set op $p
	    writeln
	    if {$snmp_verbose == "true"} {
		writeln "$hostname \[$ip\]\n$p\n"
	    }
	    writeln [format "\t%-16s = %s" \
		     "$varname\[[lindex $l 1]\]" $val]
	    incr count
	}
    }]} {
	ined acknowledge "Walking MIB tree failed for $ip."
    }
    if {$count > 0} writeln

    SnmpClose $s
}

