#! /usr/local/bin/scotty -inf
##
## Some CISCO specific proc to dump the accounting table.
##
## Copyright (c) 1993, 1994
##
## J. Schoenwaelder
## TU Braunschweig, Germany
## Institute for Operating Systems and Computer Networks
##
## Permission to use, copy, modify, and distribute this
## software and its documentation for any purpose and without
## fee is hereby granted, provided that this copyright
## notice appears in all copies.  The University of Braunschweig
## makes no representations about the suitability of this
## software for any purpose.  It is provided "as is" without
## express or implied warranty.
##

LoadDefaults snmp

SnmpInit SNMP-CISCO

##
## Write background error messages to stderr.
##

proc scottyerror { msg } { 
    global errorInfo
    writeln "$msg\n$errorInfo"
}

##
## Sort a list in numerical order. The first argument is the list
## to be sorted and the second argument gives the index of the 
## element that will be used as the key.
##

proc numcmp {field a b} {
    return [expr {[lindex $a $field] < [lindex $b $field]}]
}

proc numsort {list field} {
    lsort -command "numcmp $field" $list 
}

##
## Display a scalar variable in the log window.
##

proc ShowVariable { s variable label } {
    if {![catch {SnmpGet $s $variable.0} res]} {
	set res [split $res "\n"]
	writeln [format "  %-30s%s" $label: [lindex $res 0]]
    }
}

##
## Display the system MIB of CISCO devices.
##

proc "Cisco Information" {list} {
    set ciscomib .iso.org.dod.internet.private.enterprises.cisco
    foreach comp $list {
	if {[ined type $comp]=="NODE"} {
            set host [lindex [ined name $comp] 0]
            set ip [GetIpAddress $comp]
	    if {$ip==""} {
                ined acknowledge "Can not lookup IP Address for $host."
                continue
            }

	    set s [SnmpOpen $ip]
	    if {![catch {SnmpGet $s system.sysObjectID.0} res]} {
		set res [join $res]
		if {![string match $ciscomib* $res]} {
		    writeln "$host \[$ip\] is not a CISCO device."
		    writeln
		    SnmpClose $s
		    continue
		}
            }
	    
	    snmp default $s prefix $ciscomib.local.lsystem
	    writeln "Cisco system information for $host \[$ip\]:"
	    ShowVariable $s hostName "Host name"
	    ShowVariable $s domainName "Domain name"
	    ShowVariable $s busyPer "Percent busy (5 secs)"
	    ShowVariable $s freeMem "Free memory"

	    snmp default $s prefix $ciscomib.local.lip
	    ShowVariable $s actAge "Age of accounting data"
	    ShowVariable $s ckactAge "Age of saved accounting data"

	    SnmpClose $s
	    writeln
	}
    }
}

##
## Dump the accounting table of the cisco MIB.
##

proc "Accounting Table" {list} {
    set ciscomib .iso.org.dod.internet.private.enterprises.cisco
    foreach comp $list {
        if {[ined type $comp] == "NODE"} {
            set host [lindex [ined name $comp] 0]
            set ip [GetIpAddress $comp]
            if {$ip == ""} {
                ined acknowledge "Can not lookup IP Address for $host."
                continue;
            }
            set s [SnmpOpen $ip]
	    if {[catch {SnmpGet $s $ciscomib.local.lip.actAge.0} age]} {
		ined acknowledge "Can not get accounting information of $host."
		continue
	    }
            snmp default $s prefix \
                "$ciscomib.local.lip.lipAccountingTable.lipAccountEntry"
	    writeln "CISCO accounting statistics of $host ([join $age]):"
	    set table ""
	    catch {
		SnmpGetBulk $s {
		    actSrc
		    actDst
		    actPkts
		    actByts
		} row {
		    append table [join $row]
		}
	    } error
	    if {$error != ""} {
		ined acknowledge "Reading accounting table failed for $host."
	    } else {
		set table [numsort $table 3]
		writeln "Source       Destination      Packets      Bytes"
		foreach row $table {
		    writeln [eval "format {%-16s %-16s %16s %16s} $row"]
		}
	    }
	    SnmpClose $s
	    writeln
        }
    }
}

proc "Checkpoint Accounting Table" {list} {
    set ciscomib .iso.org.dod.internet.private.enterprises.cisco
    foreach comp $list {
        if {[ined type $comp] == "NODE"} {
            set host [lindex [ined name $comp] 0]
            set ip [GetIpAddress $comp]
            if {$ip == ""} {
                ined acknowledge "Can not lookup IP Address for $host."
                continue;
            }
            set s [SnmpOpen $ip]
	    if {[catch {SnmpGet $s $ciscomib.local.lip.ckactAge.0} age]} {
		ined acknowledge "Can not get accounting information of $host."
		continue
	    }
            snmp default $s prefix \
                "$ciscomib.local.lip.lipCkAccountingTable.lipCkAccountEntry"
	    writeln "CISCO checkpoint accounting statistics of $host ([join $age]):"
	    set table ""
	    catch {
		SnmpGetBulk $s {
		    ckactSrc
		    ckactDst
		    ckactPkts
		    ckactByts
		} row {
		    append table [join $row]
		}
	    } error
	    if {$error != ""} {
		ined acknowledge "Reading accounting table failed for $host."
	    } else {
		set table [numsort $table 3]
		writeln "Source       Destination      Packets      Bytes"
		foreach row $table {
		    writeln [eval "format {%-16s %-16s %16s %16s} $row"]
		}
	    }
	    SnmpClose $s
	    writeln
        }
    }
}

##
##
##

proc "Bytes send from Host" {list} {

    static src

    if {![info exists src]} {
	set src ""
    }

    set res [ined request "Source IP address:" [list [list "" $src]] \
	       [list cancel start] ]
    if {[lindex $res 0] == "cancel"} return
    set res [lindex $res 1]

    # should convert hostnames to ip dot notation

    if {![regexp "^\[0-9\]+\.\[0-9\]+\.\[0-9\]+\.\[0-9\]+$" $res]} {
	if {[catch {nslook $res} res]} {
	    ined acknowledge "Can not convert to IP address."
	    return
	}
    }
    set src $res

    set ciscomib .iso.org.dod.internet.private.enterprises.cisco
    foreach comp $list {
        if {[ined type $comp] == "NODE"} {
            set host [lindex [ined name $comp] 0]
            set ip [GetIpAddress $comp]
            if {$ip == ""} {
                ined acknowledge "Can not lookup IP Address for $host."
                continue;
            }
            set s [SnmpOpen $ip]
	    if {[catch {SnmpGet $s $ciscomib.local.lip.actAge.0} age]} {
		ined acknowledge "Can not get accounting information of $host."
		continue
	    }
            snmp default $s prefix \
                "$ciscomib.local.lip.lipAccountingTable.lipAccountEntry"

	    set traffic ""
	    SnmpGetBulk $s {actSrc.$src actDst.$src actPkts.$src actByts.$src} x {
		set x [join $x "\n"]
		lappend traffic $x
	    }

	    writeln [format "%16s %16s %12s %12s" \
		     Source Destination Packets Bytes]
	    writeln \
		"------------------------------------------------------------"
	    foreach x [numsort $traffic 3] {
		set src  [lindex $x 0]
		set dst  [lindex $x 1]
		set pkts [lindex $x 2]
		set byts [lindex $x 3]
		write [format "%16s " $src]
		write [format "%16s" $dst] "showName $dst $pkts $byts"
		writeln [format " %12s %12s" $pkts $byts]
	    }

	    SnmpClose $s
            writeln
	}
    }
}

proc showName { dst pkts byts } {

    static x y

    if {![info exists x]} { set x 50 }
    if {![info exists y]} { set y 50 }

    set name ""
    if {[catch {dns ptr $dst} name]} {
	if {[catch {nslook $dst} name]} {
	    set result [ined confirm "Can not look up name for $dst." "" \
			"Create a NODE for it?" [list no yes] ]
	    if {$result != "yes"} return
	}
    }

    if {$name != ""} {
	set result [ined confirm "$name \[$dst\]" "" "Create a NODE for it?" \
		   [list no yes] ]
    }

    if {$result == "yes"} {
	set id [ined create NODE]
	ined -noupdate name $id $name
	ined -noupdate address $id $dst
	ined -noupdate atttext $id text "$dst $pkts pkts $byts byte"
	ined -noupdate label $id text
	ined -noupdate move $id [incr x 20] $y
	if {[incr y 20] > 500} { set y 50 }
    }
}

##
## Set the parameters (community, timeout, retry) for snmp requests.
##

proc "Set Parameter" {list} {
    SnmpParameter
}

##
## Show the defaults as loaded from the tkined.defaults files.
##

proc "Show Defaults" {list} {
    ShowDefaults
}

##
## Delete the menus created by this interpreter.
##

proc "Delete SNMP-CISCO" {list} {
    global menus
    foreach id $menus { ined delete $id }
    exit
}

##
## Display some help about this tool.
##

proc "Help SNMP-CISCO" {list} {
    ined browse "Help about SNMP-CISCO" {
	"Cisco Information:" 
	"    Display some interesting system specific information for a" 
	"    CISCO device." 
	"" 
	"Set Parameter:" 
	"    This dialog allows you to set SNMP parameters like retries, " 
	"    timeouts, community name and port number. " 
	"" 
	"Show Defaults:" 
	"    Show the defaults that may be defined in the tkined.defaults" 
	"    files. This script makes use of definitions in the form" 
	"" 
	"        snmp.community:      <string>" 
	"        snmp.port:           <number>" 
	"        snmp.retries:        <number>" 
	"        snmp.timeout:        <number>" 
	"" 
	"        snmp.community.<ip>: <string>" 
	"        snmp.port.<ip>:      <number>" 
	"        snmp.retries.<ip>:   <number>" 
	"        snmp.timeout.<ip>:   <number>" 
	"" 
	"    where <ip> is an IP address in dot notation." 
    }
}

set menus [ ined create MENU "SNMP-CISCO" "Cisco Information" "" \
	    "Accounting Table" "Checkpoint Accounting Table" \
	    "Bytes send from Host" "" \
	    "Set Parameter" "Show Defaults" "" \
	    "Help SNMP-CISCO" "Delete SNMP-CISCO"]
